// FSyncExperiment.cpp : Defines the entry point for the console application.
//

#include "windows.h"

#include <stdio.h>
#include <tchar.h>
#include <string>
#include <iostream>

static HANDLE
open_file(const std::string& path)
{
    return CreateFile(path.c_str(), GENERIC_READ | GENERIC_WRITE, 0,
                      NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
}

static bool
write_file(HANDLE file,
           const std::string &data)
{
    const void *buffer = data.c_str();
    DWORD size = data.length();
    DWORD written;

    return WriteFile(file, buffer, size, &written, NULL) && written == size;
}

static std::string
make_buffer(const std::string& run)
{
    std::string buffer = "run number " + run + "\r\n";
    for (int i = 0; i < 10; ++i)
        buffer += buffer;

    return buffer;
}

static bool
fsync(const std::string& path)
{
    HANDLE file = open_file(path);
    if (file == INVALID_HANDLE_VALUE)
        return false;
    if (!FlushFileBuffers(file))
        return false;

    CloseHandle(file);

    return true;
}

static bool
reopen_fsync_run(const std::string & parent_path,
	             const std::string& run)
{
	std::string common_path = parent_path + "\\reopen_fsync_" + run;
	std::string fsync_path = common_path + ".sync";
	std::string nosync_path = common_path + ".nosync";

    std::string buffer = make_buffer(run);

    HANDLE nosync_file = open_file(nosync_path);
    if (nosync_file == INVALID_HANDLE_VALUE)
        return false;

    HANDLE fsync_file = open_file(fsync_path);
    if (fsync_file == INVALID_HANDLE_VALUE)
        return false;

    for (int i = 0; i < 100; ++i)
	{
        if (!write_file(nosync_file, buffer))
            return false;
        if (!write_file(fsync_file, buffer))
            return false;
	}

    if (!CloseHandle(nosync_file))
        return false;
    if (!CloseHandle(fsync_file))
        return false;

    return fsync(fsync_path);
}

static bool
rename_run(const std::string & parent_path,
           const std::string& run,
           bool fsync_after_rename)
{
    std::string common_path = parent_path + "\\rename_" + run;
    std::string temp_path = common_path + ".temp";
    std::string final_path = common_path + ".final";

    std::string buffer = make_buffer(run);

    HANDLE temp_file = open_file(temp_path);
    if (temp_file == INVALID_HANDLE_VALUE)
        return false;
    if (!write_file(temp_file, buffer))
        return false;
    if (!FlushFileBuffers(temp_file))
        return false;
    if (!CloseHandle(temp_file))
        return false;

    if (!MoveFile(temp_path.c_str(), final_path.c_str()))
        return false;

    if (fsync_after_rename)
        return fsync(final_path);

    return true;
}

static void
test_reopen_fsync(const std::string& path)
{
    for (int run = 1; reopen_fsync_run(path, std::to_string(run)); ++run)
        std::cout << "Last fsync'ed run: " << run << std::endl;
}

static void
test_rename(const std::string& path)
{
    for (int run = 1; rename_run(path, std::to_string(run), false); ++run)
        std::cout << "Last complete rename run: " << run << std::endl;
}

static void
test_rename_fsync(const std::string& path)
{
    for (int run = 1; rename_run(path, std::to_string(run), true); ++run)
        std::cout << "Last fsync'ed rename run: " << run << std::endl;
}

int _tmain(int argc, _TCHAR* argv[])
{
    const std::string base_path = "E:";
 
//    test_reopen_fsync(base_path + "\\reopen_fsync");
//    test_rename(base_path + "\\rename");
    test_rename_fsync(base_path + "\\rename_fsync");

    return 0;
}

